#include <joyos.h>
#include <happylib.h>
#include <math.h>
#include "dmath.h"
#include "sensing.h"
#include <lib/pid.h>

#define L_GAIN (1.0)
#define R_GAIN (1.0)
//never set gains higher than 1!!!
//right wheel should be on 1, left wheel on 0

//the distance between the driving wheels
#define WIDTH (8.82)
#define PAUSE (12)
#define MIN_VEL (50)
#define PI (3.1415)

#define angleDist(f, t) (makeCircular(t-f))

/**
Note: it may be desireable to modify the relative gains of the left and right motors. To tune the gains,
uncomment the lines above containing 'frob_read' and comment out the lines below them. By doing this it 
is possible to use the frob knob to tune the parameters.
*/

//struct pid_controller headingPID;

//sets the left motor velocity
void set_left(int vel){
  motor_set_vel(0, clip(L_GAIN*vel, -255, 255));
}

//sets the right motor velocity
void set_right(int vel){
  motor_set_vel(1, clip(R_GAIN*vel, -255, 255));
}


//sets both left and right wheel motors
void setPair(int left, int right){
  set_left(left);
  set_right(right);
}

//makes a turn towards
//target: desired angle
//tol: tolerance
//gain: how fast to turn
//timeout: seconds allowed
//returns success
bool proportionalTurnTo(float target, float tol, float gain, float timeout){
	long start = get_time();
	bool success = true;
	int t = makeCircular(target);
	copy_objects();
	int err = angleDist(getHeading(), t);
	while ( abs(err) > tol ){
		int step = (int)(err*gain);
		if (get_time() - start > timeout * 1000){
			success = false;
			break;
		}
		if (abs(step) < MIN_VEL){
			gain *= 1.6;
		}
		setPair( -step, step);
		copy_objects();
		err = angleDist(getHeading(), t);
	}
	setPair(0,0);
	pause(PAUSE);
	return success;
}

//amount: how much to turn by. The other variables are as above
bool proportionalTurnBy(float amount, float tol, float gain, float timeout){
	return proportionalTurnTo( amount + getHeading(), tol, gain, timeout);
}

//the robot turs so that it faces p
bool alignWith(Point* p, float gain, int tol, float timeout){
	copy_objects();
	float turn = cleverAngleOf(p->x - myX(), p->y - myY() );
	//float turn = angleOf(p->x - myX(), p->y - myY() );
	bool b = proportionalTurnTo(turn, tol, gain, timeout);
	return b;
}

bool radianDrive(Point* p, float tol, float g1, float g2, float timeout){
	long start = get_time();
	bool success = true;
	copy_objects();
	Point now = {.x = myX(), .y = myY()};
	float head = getHeading();
	double err = distanceBetweenPoints(p,&now);
	while (err > tol){
		if (get_time() - start > timeout * 1000){
			success = false;
			break;
		}
		
		float rm = sqrt(now.x*now.x + now.y*now.y);
		float rg = sqrt(p->x*p->x + p->y*p->y);
		float dt = (rm+rg)*angleDist(angleOf(now.x,now.y), angleOf(p->x,p->y))/57.29578/2;
		err = sqrt( (rm-rg)*(rm-rg) + dt*dt );
		
		int fvel = clip(g1*err,-255,255);
		int rvel = angleDist(head, angleOf(rg-rm, dt) + angleOf(now.x,now.y));
		if (abs(rvel) > 90){
			fvel = -fvel;
			rvel = makeCircular(2*rvel)/2;
		}
		fvel = (int)(fvel*pow(cos(rvel/57.29578), 10));
		rvel = clip(g2*rvel,-255,255);
		setPair(fvel - rvel, fvel + rvel);
		
		//last = position_microtime;
		copy_objects();
		now.x = myX();
		now.y = myY();
		head = getHeading();
	}
	setPair(0,0);
	pause(PAUSE);
	return success;
}



void persistentDrive(Point* p, float tol){
	while(!radianDrive(p, tol, 17, 3.5, 6)){
		pause(500);
		//stuck on opponent !!improve this
		float f = angleDist(getHeading(), angleOf(myX(),myY()));
		if (distanceBetween(myX(), myY(), oppX(), oppY()) < 15){
			int vel = abs(angleDist( getHeading(),angleOf(oppX()-myX(),oppY()-myY()))) < 90 ? 200 : -200;
			setPair(-vel, -vel);
			pause(1500);
			setPair(0,0);
			proportionalTurnBy(45, 10, 5, 2);
		}
		//stuck on inner wall
		else if (distanceBetween(0, 0, myX(), myY()) < 20){
			if (abs(f) > 90)
				setPair(-200,-200);
			else
				setPair(200,200);
			pause(700);
			proportionalTurnBy(40, 10, 5, 2);
		}
		//stuck on outer wall
		else if (distanceBetween(0, 0, myX(), myY()) > 40){
			//gear box or lever
			if (abs(f) > 90)
				setPair(200,200);
			else
				setPair(-200,-200);
			pause(1000);
			setPair(0,0);
			proportionalTurnBy(30, 10, 5, 2);
		}
		else{
			setPair(200,200);
			pause(500);
			setPair(0,0);
			proportionalTurnBy(30, 10, 5, 2);
		}
		calibrate();
	}
}
